
/*

  rsa.c

  Author: Mika Kojo <mkojo@ssh.fi>

  Copyright (C) 1997 SSH Communications Security Oy, Espoo, Finland
  All rights reserved.

  Created: Thu Jun 12 17:40:09 1997 [mkojo]

  Alternate take on RSA, modified after Tatu Ylonen's original SSH
  implementation.

  */

/*

rsa.c

Author: Tatu Ylonen <ylo@cs.hut.fi>

Copyright (c) 1995 Tatu Ylonen <ylo@cs.hut.fi>, Espoo, Finland
                   All rights reserved

Created: Fri Mar  3 22:07:06 1995 ylo

Description of the RSA algorithm can be found e.g. from the following sources:

  Bruce Schneier: Applied Cryptography.  John Wiley & Sons, 1994.

  Jennifer Seberry and Josed Pieprzyk: Cryptography: An Introduction to
    Computer Security.  Prentice-Hall, 1989.

  Man Young Rhee: Cryptography and Secure Data Communications.  McGraw-Hill,
    1994.

  R. Rivest, A. Shamir, and L. M. Adleman: Cryptographic Communications
    System and Method.  US Patent 4,405,829, 1983.

  Hans Riesel: Prime Numbers and Computer Methods for Factorization.
    Birkhauser, 1994.

  The RSA Frequently Asked Questions document by RSA Data Security, Inc., 1995.

  RSA in 3 lines of perl by Adam Back <aba@atlax.ex.ac.uk>, 1995, as included
    below:
    #!/usr/local/bin/perl -s-- -export-a-crypto-system-sig -RSA-in-3-lines-PERL
    ($k,$n)=@ARGV;$m=unpack(H.$w,$m."\0"x$w),$_=`echo "16do$w 2+4Oi0$d*-^1[d2%
    Sa2/d0<X+d*La1=z\U$n%0]SX$k"[$m*]\EszlXx++p|dc`,s/^.|\W//g,print pack('H*'
    ,$_)while read(STDIN,$m,($w=2*$d-1+length($n||die"$0 [-d] k n\n")&~1)/2)

*/

/*
 * $Id: rsa.c,v 1.62 2001/04/24 18:29:43 fis Exp $
 * $Log: rsa.c,v $
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 *
 * $EndLog$
 */

#include "sshincludes.h"
#include "sshmp.h" /* was "gmp.h" */
#include "sshcrypt.h"
#include "sshcrypti.h"
#include "genmp.h"
#include "sshgetput.h"
#include "sshencode.h"
#include "rsa.h"
#include "sshasn1.h"


#ifdef WITH_RSA






#define SSH_RSA_MINIMUM_PADDING 10
#define SSH_RSA_MAX_BYTES       65535








typedef struct
{
  size_t bits;
  SshMPIntStruct p, q, e, d, n, u;
} SshRSAPrivateKey;

typedef struct
{
  size_t bits;
  SshMPIntStruct n, e;
} SshRSAPublicKey;

/* Given mutual primes p and q, derives RSA key components n, d, e, and u.
   The exponent e will be at least ebits bits in size. p must be smaller
   than q.
 */

static void derive_rsa_keys(SshMPInt n, SshMPInt e, SshMPInt d, SshMPInt u,
                            SshMPInt p, SshMPInt q,
                            unsigned int ebits)
{
  SshMPIntStruct p_minus_1, q_minus_1, aux, phi, G, F;

  /* Initialize. */
  ssh_mp_init(&p_minus_1);
  ssh_mp_init(&q_minus_1);
  ssh_mp_init(&aux);
  ssh_mp_init(&phi);
  ssh_mp_init(&G);
  ssh_mp_init(&F);

  /* Compute p-1 and q-1. */
  ssh_mp_sub_ui(&p_minus_1, p, 1);
  ssh_mp_sub_ui(&q_minus_1, q, 1);

  /* phi = (p - 1) * (q - 1); the number of positive integers less than p*q
     that are relatively prime to p*q. */
  ssh_mp_mul(&phi, &p_minus_1, &q_minus_1);

  /* G is the number of "spare key sets" for a given modulus n.  The smaller
     G is, the better.  The smallest G can get is 2. This tells
     in practice nothing about the safety of primes p and q. */
  ssh_mp_gcd(&G, &p_minus_1, &q_minus_1);

  /* F = phi / G; the number of relative prime numbers per spare key set. */
  ssh_mp_div_q(&F, &phi, &G);

  /* Find a suitable e (the public exponent). */
  ssh_mp_set_ui(e, 1);
  ssh_mp_mul_2exp(e, e, ebits);
  ssh_mp_sub_ui(e, e, 1); /* make lowest bit 1, and substract 2. */
  /* Keep adding 2 until it is relatively prime to (p-1)(q-1). */
  do
    {
      ssh_mp_add_ui(e, e, 2);
      ssh_mp_gcd(&aux, e, &phi);
    }
  while (ssh_mp_cmp_ui(&aux, 1) != 0);

  /* d is the multiplicative inverse of e, mod F.  Could also be mod
     (p-1)(q-1); however, we try to choose the smallest possible d. */
  ssh_mp_mod_invert(d, e, &F);

  /* u is the multiplicative inverse of p, mod q, if p < q.  It is used
     when doing private key RSA operations using the chinese remainder
     theorem method. */
  ssh_mp_mod_invert(u, p, q);

  /* n = p * q (the public modulus). */
  ssh_mp_mul(n, p, q);

  /* Clear auxiliary variables. */
  ssh_mp_clear(&p_minus_1);
  ssh_mp_clear(&q_minus_1);
  ssh_mp_clear(&aux);
  ssh_mp_clear(&phi);
  ssh_mp_clear(&G);
  ssh_mp_clear(&F);
}

/* Generate RSA keys with e set to a fixed value. */
static int derive_rsa_keys_with_e(SshMPInt n, SshMPInt e, SshMPInt d,
                                  SshMPInt u, SshMPInt p, SshMPInt q,
                                  SshMPInt given_e)
{
  SshMPIntStruct p_minus_1, q_minus_1, aux, phi, G, F;
  int rv;

  /* Initialize. */
  ssh_mp_init(&p_minus_1);
  ssh_mp_init(&q_minus_1);
  ssh_mp_init(&aux);
  ssh_mp_init(&phi);
  ssh_mp_init(&G);
  ssh_mp_init(&F);

  /* Compute p-1 and q-1. */
  ssh_mp_sub_ui(&p_minus_1, p, 1);
  ssh_mp_sub_ui(&q_minus_1, q, 1);

  /* phi = (p - 1) * (q - 1); the number of positive integers less than p*q
     that are relatively prime to p*q. */
  ssh_mp_mul(&phi, &p_minus_1, &q_minus_1);

  /* G is the number of "spare key sets" for a given modulus n.  The smaller
     G is, the better.  The smallest G can get is 2. This tells
     in practice nothing about the safety of primes p and q. */
  ssh_mp_gcd(&G, &p_minus_1, &q_minus_1);

  /* F = phi / G; the number of relative prime numbers per spare key set. */
  ssh_mp_div_q(&F, &phi, &G);

  /* Find a suitable e (the public exponent). */
  ssh_mp_set(e, given_e);
  if (ssh_mp_cmp_ui(e, 3) < 0)
    {
      rv = 0;
      goto failed;
    }

  /* Transform the e into something that is has some probability of
     being correct. */
  if ((ssh_mp_get_ui(e) & 0x1) == 0)
    ssh_mp_add_ui(e, e, 1);
  ssh_mp_sub_ui(e, e, 2);
  /* Keep adding 2 until it is relatively prime to (p-1)(q-1). */
  do
    {
      ssh_mp_add_ui(e, e, 2);
      ssh_mp_gcd(&aux, e, &phi);
    }
  while (ssh_mp_cmp_ui(&aux, 1) != 0);

  /* Verify that the e is correct still! */
  if (ssh_mp_cmp(e, given_e) != 0)
    rv = 1;
  else
    rv = 2;

  /* d is the multiplicative inverse of e, mod F.  Could also be mod
     (p-1)(q-1); however, we try to choose the smallest possible d. */
  ssh_mp_mod_invert(d, e, &F);

  /* u is the multiplicative inverse of p, mod q, if p < q.  It is used
     when doing private key RSA operations using the chinese remainder
     theorem method. */
  ssh_mp_mod_invert(u, p, q);

  /* n = p * q (the public modulus). */
  ssh_mp_mul(n, p, q);

failed:

  /* Clear auxiliary variables. */
  ssh_mp_clear(&p_minus_1);
  ssh_mp_clear(&q_minus_1);
  ssh_mp_clear(&aux);
  ssh_mp_clear(&phi);
  ssh_mp_clear(&G);
  ssh_mp_clear(&F);

  return rv;
}

void *ssh_rsa_make_private_key_of_all(SshMPInt p, SshMPInt q, SshMPInt n,
                                      SshMPInt e, SshMPInt d, SshMPInt u)
{
  SshRSAPrivateKey *private_key = ssh_xmalloc(sizeof(*private_key));

  ssh_mp_init_set(&private_key->e, e);
  ssh_mp_init_set(&private_key->d, d);
  ssh_mp_init_set(&private_key->n, n);
  ssh_mp_init_set(&private_key->u, u);
  ssh_mp_init_set(&private_key->p, p);
  ssh_mp_init_set(&private_key->q, q);
  private_key->bits = ssh_mp_bit_size(n);

  return (void *)private_key;
}

void ssh_rsa_private_key_init(SshRSAPrivateKey *private_key)
{
  /* Initialize with zeroes. */
  ssh_mp_init_set_ui(&private_key->e, 0);
  ssh_mp_init_set_ui(&private_key->d, 0);
  ssh_mp_init_set_ui(&private_key->n, 0);
  ssh_mp_init_set_ui(&private_key->u, 0);
  ssh_mp_init_set_ui(&private_key->p, 0);
  ssh_mp_init_set_ui(&private_key->q, 0);
  private_key->bits = 0;
}

/* Almost same as above but is given d also. Creates the valid
   SshRSAPrivateKey. Is used from action make routines. */
void *ssh_rsa_make_private_key_of_pqd(SshMPInt p, SshMPInt q, SshMPInt d)
{
  SshMPIntStruct p_minus_1, q_minus_1, aux, phi, G, F;
  SshRSAPrivateKey *private_key = ssh_xmalloc(sizeof(*private_key));

  /* Initialize. */
  ssh_mp_init(&p_minus_1);
  ssh_mp_init(&q_minus_1);
  ssh_mp_init(&aux);
  ssh_mp_init(&phi);
  ssh_mp_init(&G);
  ssh_mp_init(&F);

  /* Initialize the private key. */
  ssh_mp_init(&private_key->e);
  ssh_mp_init(&private_key->d);
  ssh_mp_init(&private_key->u);
  ssh_mp_init(&private_key->n);
  ssh_mp_init(&private_key->p);
  ssh_mp_init(&private_key->q);

  /* Compute p-1 and q-1. */
  ssh_mp_sub_ui(&p_minus_1, p, 1);
  ssh_mp_sub_ui(&q_minus_1, q, 1);

  /* Set the p and q. */
  ssh_mp_set(&private_key->p, p);
  ssh_mp_set(&private_key->q, q);

  /* phi = (p - 1) * (q - 1); the number of positive integers less than p*q
     that are relatively prime to p*q. */
  ssh_mp_mul(&phi, &p_minus_1, &q_minus_1);

  /* G is the number of "spare key sets" for a given modulus n.  The smaller
     G is, the better.  The smallest G can get is 2. This tells
     in practice nothing about the safety of primes p and q. */
  ssh_mp_gcd(&G, &p_minus_1, &q_minus_1);

  /* F = phi / G; the number of relative prime numbers per spare key set. */
  ssh_mp_div_q(&F, &phi, &G);

  /* Find a suitable e (the public exponent). */
  ssh_mp_mod_invert(&private_key->e, d, &phi);
  ssh_mp_set(&private_key->d, d);

  /* u is the multiplicative inverse of p, mod q, if p < q.  It is used
     when doing private key RSA operations using the chinese remainder
     theorem method. */
  ssh_mp_mod_invert(&private_key->u, p, q);

  /* n = p * q (the public modulus). */
  ssh_mp_mul(&private_key->n, p, q);

  /* Clear auxiliary variables. */
  ssh_mp_clear(&p_minus_1);
  ssh_mp_clear(&q_minus_1);
  ssh_mp_clear(&aux);
  ssh_mp_clear(&phi);
  ssh_mp_clear(&G);
  ssh_mp_clear(&F);

  /* Compute the bit size of the key. */
  private_key->bits = ssh_mp_bit_size(&private_key->n);

  /* Return the private key object. */
  return (void *)private_key;
}

/* Generates RSA public and private keys.  This initializes the data
   structures; they should be freed with rsa_clear_private_key and
   rsa_clear_public_key. */

void *ssh_rsa_generate_private_key(unsigned int bits,
                                   SshMPInt e)
{
  SshRSAPrivateKey *prv = ssh_xmalloc(sizeof(*prv));
  SshMPIntStruct test, aux, min, max;
  unsigned int pbits;
#if SSH_USE_OLD_RSA_MODULUS_GENERATION
  unsigned int qbits;
#endif
  int ret;

  /* Initialize our key. */
  ssh_mp_init(&prv->q);
  ssh_mp_init(&prv->p);
  ssh_mp_init(&prv->e);
  ssh_mp_init(&prv->d);
  ssh_mp_init(&prv->u);
  ssh_mp_init(&prv->n);

  /* Auxliary variables. */
  ssh_mp_init(&test);
  ssh_mp_init(&aux);
  ssh_mp_init(&min);
  ssh_mp_init(&max);

  /* Compute the number of bits in each prime. */
  pbits = bits / 2;

  /* Generate random number p. */
  ssh_mp_random_prime(&prv->p, pbits);

  /* Repeat until one finds primes that are distinct. */
retry:

#if SSH_USE_OLD_RSA_MODULUS_GENERATION
  qbits = bits - pbits;
  /* Generate random number q. */
  ssh_mp_random_prime(&prv->q, qbits);
#else /* SSH_USE_OLD_RSA_MODULUS_GENERATION */
  /* Form 2^(bits). */
  ssh_mp_set_ui(&aux, 0);
  ssh_mp_set_bit(&aux, bits);
  /* Divide to get the maximum. */
  ssh_mp_div_q(&max, &aux, &prv->p);
  /* Form 2^(bits-1) - 1 (actually 2^(bits-1) for the same effort). */
  ssh_mp_set_ui(&aux, 0);
  ssh_mp_set_bit(&aux, bits-1);
  /* Divide to get the minimum. */
  ssh_mp_div_q(&min, &aux, &prv->p);

  /* Generate a prime number between (min, max). That is,
     we get min < q < max, and then

       min * p < 2^(bits-1) < q * p < max * p < 2^(bits),

     as we desire. The higher bound is trivial, but the lower bound
     follows as min * p < 2^(bits-1) < (min + 1) * p.
     */

  ssh_mp_random_prime_within_interval(&prv->q, &min, &max);
#endif /* SSH_USE_OLD_RSA_MODULUS_GENERATION */

  /* Sort them so that p < q. */
  ret = ssh_mp_cmp(&prv->p, &prv->q);
  if (ret == 0)
    goto retry;

  if (ret > 0)
    {
      ssh_mp_set(&aux, &prv->p);
      ssh_mp_set(&prv->p, &prv->q);
      ssh_mp_set(&prv->q, &aux);
    }

  /* Make certain p and q are relatively prime (in case one or both were false
     positives...  Though this is quite impossible). */
  ssh_mp_gcd(&aux, &prv->p, &prv->q);
  if (ssh_mp_cmp_ui(&aux, 1) != 0)
    goto retry;

  if (e == NULL)
    /* Derive the RSA private key from the primes. */
    derive_rsa_keys(&prv->n, &prv->e, &prv->d, &prv->u, &prv->p, &prv->q,
                    16);
  else
    switch (derive_rsa_keys_with_e(&prv->n, &prv->e, &prv->d, &prv->u,
                                   &prv->p, &prv->q,
                                   e))
      {
      case 0:
        ssh_xfree(prv);
        ssh_mp_clear(&aux);
        ssh_mp_clear(&test);
        ssh_mp_clear(&min);
        ssh_mp_clear(&max);
        return NULL;
      case 1:
      case 2:
        /* Do nothing special, accept possible changes to the given value. */
        break;
      }

  ssh_mp_clear(&aux);
  ssh_mp_clear(&test);
  ssh_mp_clear(&min);
  ssh_mp_clear(&max);

  /* Compute the bit size of the key. */
  prv->bits = ssh_mp_bit_size(&prv->n);
  return (void *)prv;
}

/* Methods for handling private key generation, from the generic
   interface. */

typedef struct
{
  unsigned int bits;
  SshMPIntStruct n, p, q, e, d, u;
} SshRSAInitCtx;

/* Initialization functions. */

void *ssh_rsa_private_key_init_action()
{
  SshRSAInitCtx *ctx = ssh_xmalloc(sizeof(*ctx));

  ssh_mp_init_set_ui(&ctx->n, 0);
  ssh_mp_init_set_ui(&ctx->p, 0);
  ssh_mp_init_set_ui(&ctx->q, 0);
  ssh_mp_init_set_ui(&ctx->e, 0);
  ssh_mp_init_set_ui(&ctx->d, 0);
  ssh_mp_init_set_ui(&ctx->u, 0);

  ctx->bits = 0;

  return (void *)ctx;
}

void *ssh_rsa_public_key_init_action(void)
{
  return ssh_rsa_private_key_init_action();
}

void ssh_rsa_private_key_init_ctx_free(void *context)
{
  SshRSAInitCtx *ctx = context;

  ssh_mp_clear(&ctx->n);
  ssh_mp_clear(&ctx->p);
  ssh_mp_clear(&ctx->q);
  ssh_mp_clear(&ctx->e);
  ssh_mp_clear(&ctx->d);
  ssh_mp_clear(&ctx->u);

  ssh_xfree(ctx);
}

/* Special actions. */

char *ssh_rsa_action_put(void *context, va_list ap,
                         void *input_context,
                         SshCryptoType type,
                         SshPkFormat format)
{
  SshRSAInitCtx *ctx = context;
  SshMPInt temp;
  char *r;

  r = "p";
  switch (format)
    {
    case SSH_PKF_SIZE:
      if (type & SSH_CRYPTO_TYPE_PUBLIC_KEY)
        return NULL;
      ctx->bits = va_arg(ap, unsigned int);
      r = "i";
      break;
    case SSH_PKF_PRIME_P:
      if (type & SSH_CRYPTO_TYPE_PUBLIC_KEY)
        return NULL;
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(&ctx->p, temp);
      break;
    case SSH_PKF_PRIME_Q:
      if (type & SSH_CRYPTO_TYPE_PUBLIC_KEY)
        return NULL;
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(&ctx->q, temp);
      break;
    case SSH_PKF_MODULO_N:
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(&ctx->n, temp);
      break;
    case SSH_PKF_SECRET_D:
      if (type & SSH_CRYPTO_TYPE_PUBLIC_KEY)
        return NULL;
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(&ctx->d, temp);
      break;
    case SSH_PKF_INVERSE_U:
      if (type & SSH_CRYPTO_TYPE_PUBLIC_KEY)
        return NULL;
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(&ctx->u, temp);
      break;
    case SSH_PKF_PUBLIC_E:
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(&ctx->e, temp);
      break;
    default:
      return NULL;
    }
  return r;
}

char *ssh_rsa_action_private_key_put(void *context, va_list ap,
                                     void *input_context,
                                     SshPkFormat format)
{
  return ssh_rsa_action_put(context, ap,
                            input_context,
                            SSH_CRYPTO_TYPE_PRIVATE_KEY,
                            format);
}

char *ssh_rsa_action_private_key_get(void *context, va_list ap,
                                     void **output_context,
                                     SshPkFormat format)
{
  SshRSAPrivateKey *prv = context;
  unsigned int *size;
  SshMPInt temp;
  char *r;

  r = "p";
  switch (format)
    {
    case SSH_PKF_SIZE:
      size = va_arg(ap, unsigned int *);
      *size = ssh_mp_bit_size(&prv->n);
      break;
    case SSH_PKF_PRIME_P:
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(temp, &prv->p);
      break;
    case SSH_PKF_PRIME_Q:
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(temp, &prv->q);
      break;
    case SSH_PKF_MODULO_N:
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(temp, &prv->n);
      break;
    case SSH_PKF_SECRET_D:
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(temp, &prv->d);
      break;
    case SSH_PKF_INVERSE_U:
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(temp, &prv->u);
      break;
    case SSH_PKF_PUBLIC_E:
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(temp, &prv->e);
      break;
    default:
      return NULL;
    }
  return r;
}

char *ssh_rsa_action_public_key_put(void *context, va_list ap,
                                    void *input_context,
                                    SshPkFormat format)
{
  return ssh_rsa_action_put(context, ap,
                            input_context,
                            SSH_CRYPTO_TYPE_PUBLIC_KEY,
                            format);
}

char *ssh_rsa_action_public_key_get(void *context, va_list ap,
                                    void **output_context,
                                    SshPkFormat format)
{
  SshRSAPublicKey *pub = context;
  unsigned int *size;
  SshMPInt temp;
  char *r;

  r = "p";
  switch (format)
    {
    case SSH_PKF_SIZE:
      size = va_arg(ap, unsigned int *);
      *size = ssh_mp_bit_size(&pub->n);
      break;
    case SSH_PKF_MODULO_N:
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(temp, &pub->n);
      break;
    case SSH_PKF_PUBLIC_E:
      temp = va_arg(ap, SshMPInt );
      ssh_mp_set(temp, &pub->e);
      break;
    default:
      return NULL;
    }
  return r;
}

/* Making the key from this bunch of information. */

void *ssh_rsa_public_key_make_action(void *context)
{
  SshRSAInitCtx *ctx = context;
  SshRSAPublicKey *public_key;

  if (ssh_mp_cmp_ui(&ctx->e, 0) == 0 ||
      ssh_mp_cmp_ui(&ctx->n, 0) == 0)
    return NULL;

  public_key = ssh_xmalloc(sizeof(*public_key));
  ssh_mp_init_set(&public_key->e, &ctx->e);
  ssh_mp_init_set(&public_key->n, &ctx->n);

  /* Compute the size of the public key. */
  public_key->bits = ssh_mp_bit_size(&public_key->n);

  return public_key;
}

/* Try to handle the given data in a reasonable manner. */
void *ssh_rsa_private_key_make_action(void *context)
{
  SshRSAInitCtx *ctx = context;

  if (ssh_mp_cmp_ui(&ctx->d, 0) == 0 ||
      ssh_mp_cmp_ui(&ctx->p, 0) == 0 ||
      ssh_mp_cmp_ui(&ctx->q, 0) == 0)
    {

    /* Generate with e, p and q set. */
      if (ssh_mp_cmp_ui(&ctx->e, 0) != 0 &&
          ssh_mp_cmp_ui(&ctx->p, 0) != 0 &&
          ssh_mp_cmp_ui(&ctx->q, 0) != 0)
      {
          SshRSAPrivateKey *prv = ssh_xmalloc(sizeof(*prv));
          ssh_rsa_private_key_init(prv);
          ssh_mp_set(&prv->q, &ctx->q);
          ssh_mp_set(&prv->p, &ctx->p);
          (void) derive_rsa_keys_with_e(&prv->n, &prv->e, &prv->d, &prv->u,
                                        &prv->p, &prv->q, &ctx->e);
          /* XXX return value not checked. */
          return prv;
      }

      /* XXX Cannot generate because no predefined size exists. */
      if (ctx->bits == 0)
        return NULL;


      /* Generate with e set. */
      if (ssh_mp_cmp_ui(&ctx->e, 0) != 0)
        return (void *)ssh_rsa_generate_private_key(ctx->bits,
                                                    &ctx->e);

      /* Just generate from assigned values. */
      return (void *)ssh_rsa_generate_private_key(ctx->bits,
                                                  NULL);
    }
  else
    {
      if (ssh_mp_cmp_ui(&ctx->d, 0) != 0 &&
          ssh_mp_cmp_ui(&ctx->p, 0) != 0 &&
          ssh_mp_cmp_ui(&ctx->q, 0) != 0)
        {
          if (ssh_mp_cmp_ui(&ctx->e, 0) != 0 &&
              ssh_mp_cmp_ui(&ctx->n, 0) != 0 &&
              ssh_mp_cmp_ui(&ctx->u, 0) != 0)
            {
              return
                ssh_rsa_make_private_key_of_all(&ctx->p, &ctx->q,
                                                &ctx->n, &ctx->e,
                                                &ctx->d, &ctx->u);
            }

          return ssh_rsa_make_private_key_of_pqd(&ctx->p, &ctx->q, &ctx->d);
        }
    }
  return NULL;
}

void ssh_rsa_private_key_derive_public_key(const void *private_key,
                                           void **public_key)
{
  const SshRSAPrivateKey *prv = private_key;
  SshRSAPublicKey *pub;

  pub = ssh_xmalloc(sizeof(*pub));

  ssh_mp_init_set(&pub->n, &prv->n);
  ssh_mp_init_set(&pub->e, &prv->e);

  /* Size of the public key modulus. */
  pub->bits = prv->bits;

  *public_key = (void *)pub;
}

/* Frees any memory associated with the private key. */

void ssh_rsa_private_key_free(void *private_key)
{
  SshRSAPrivateKey *prv = private_key;
  ssh_mp_clear(&prv->n);
  ssh_mp_clear(&prv->e);
  ssh_mp_clear(&prv->d);
  ssh_mp_clear(&prv->u);
  ssh_mp_clear(&prv->p);
  ssh_mp_clear(&prv->q);

  ssh_xfree(prv);
}

/* Frees any memory associated with the public key. */

void ssh_rsa_public_key_free(void *public_key)
{
  SshRSAPublicKey *pub = public_key;

  ssh_mp_clear(&pub->e);
  ssh_mp_clear(&pub->n);

  ssh_xfree(pub);
}

/* Explicit copying routines. */

void ssh_rsa_public_key_copy(void *op_src, void **op_dest)
{
  SshRSAPublicKey *pub_dest, *pub_src = op_src;

  pub_dest = ssh_xmalloc(sizeof(*pub_dest));

  ssh_mp_init_set(&pub_dest->n, &pub_src->n);
  ssh_mp_init_set(&pub_dest->e, &pub_src->e);
  pub_dest->bits = pub_src->bits;

  *op_dest = (void *)pub_dest;
}

void ssh_rsa_private_key_copy(void *op_src, void **op_dest)
{
  SshRSAPrivateKey *prv_dest, *prv_src = op_src;

  prv_dest = ssh_xmalloc(sizeof(*prv_dest));

  /* Copy bit counts. */
  prv_dest->bits = prv_src->bits;

  *op_dest = (void *)prv_dest;

  ssh_mp_init_set(&prv_dest->n, &prv_src->n);
  ssh_mp_init_set(&prv_dest->e, &prv_src->e);
  ssh_mp_init_set(&prv_dest->d, &prv_src->d);
  ssh_mp_init_set(&prv_dest->u, &prv_src->u);
  ssh_mp_init_set(&prv_dest->p, &prv_src->p);
  ssh_mp_init_set(&prv_dest->q, &prv_src->q);
}

/* These are return values for ssh_rsa_private. */
typedef enum
{
  SSH_RSA_OP_OK,
  SSH_RSA_OP_FAILED
} SshRSAOpStatus;






























































































































/* Performs a private-key RSA operation (encrypt/decrypt).  The computation
   is done using the Chinese Remainder Theorem, which is faster than
   direct modular exponentiation. */
static SshRSAOpStatus ssh_rsa_private(SshMPInt input,
                                      SshMPInt output,
                                      const SshRSAPrivateKey *prv)
{






















  SshMPIntStruct dp, dq, p2, q2, k;

  /* Initialize temporary variables. */
  ssh_mp_init(&dp);
  ssh_mp_init(&dq);
  ssh_mp_init(&p2);
  ssh_mp_init(&q2);
  ssh_mp_init(&k);

  /* Compute dp = d mod p-1. */
  ssh_mp_sub_ui(&dp, &prv->p, 1);
  ssh_mp_mod(&dp, &prv->d, &dp);

  /* Compute dq = d mod q-1. */
  ssh_mp_sub_ui(&dq, &prv->q, 1);
  ssh_mp_mod(&dq, &prv->d, &dq);

  /* Compute p2 = (input mod p) ^ dp mod p. */
  ssh_mp_mod(&p2, input, &prv->p);
  ssh_mp_powm(&p2, &p2, &dp, &prv->p);

  /* Compute q2 = (input mod q) ^ dq mod q. */
  ssh_mp_mod(&q2, input, &prv->q);
  ssh_mp_powm(&q2, &q2, &dq, &prv->q);

  /* Compute k = ((q2 - p2) mod q) * u mod q. */
  ssh_mp_sub(&k, &q2, &p2);
  ssh_mp_mul(&k, &k, &prv->u);
  ssh_mp_mod(&k, &k, &prv->q);

  /* Compute output = p2 + p * k. */
  ssh_mp_mul(output, &prv->p, &k);
  ssh_mp_add(output, output, &p2);

  /* Clear temporary variables. */
  ssh_mp_clear(&dp);
  ssh_mp_clear(&dq);
  ssh_mp_clear(&p2);
  ssh_mp_clear(&q2);
  ssh_mp_clear(&k);
  return SSH_RSA_OP_OK;


}

/* Performs a public-key RSA operation (encrypt/decrypt). */

void ssh_rsa_public(SshMPInt input, SshMPInt output, const SshRSAPublicKey *pub)
{




















  /* Exponentiate. */
  ssh_mp_powm(output, input, &pub->e, &pub->n);


}


/* MGF1 for the OAEP. */
void ssh_rsa_mgf1(const SshHashDef *hash,
                  const unsigned char *seed, size_t seed_len,
                  unsigned char *mask, size_t mask_len)
{
  SshUInt32      i, steps;
  void          *context;
  unsigned char *digest;

  context = ssh_xmalloc((*hash->ctxsize)());
  digest  = ssh_xmalloc(hash->digest_length);

  for (i = 0, steps = 0; i < mask_len; i += hash->digest_length, steps++)
    {
      unsigned char counter[4];
      size_t avail;

      SSH_PUT_32BIT(counter, steps);

      (*hash->reset_context)(context);
      (*hash->update)(context, seed, seed_len);
      (*hash->update)(context, counter, 4);
      (*hash->final)(context, digest);

      /* Now copy the digest to the mask. */
      avail = mask_len - i;
      if (avail >= hash->digest_length)
        memcpy(mask + i, digest, hash->digest_length);
      else
        memcpy(mask + i, digest, avail);
    }

  ssh_xfree(context);
  ssh_xfree(digest);
}

/* OAEP encode.

   Note: in future a suitable library of the mask generation functions
   should be written.
 */
Boolean ssh_rsa_oaep_encode_with_mgf1(const SshHashDef *hash,
                                      const unsigned char *msg,
                                      size_t msg_len,
                                      const unsigned char *param,
                                      size_t param_len,
                                      unsigned char *emsg, size_t emsg_len)
{
  void *context;
  unsigned char *db;
  /* unsigned char *db_mask; */
  unsigned char *seed;
  size_t db_len, i;

  /* Check that the size constraints are satisfied. */
  if (msg_len > emsg_len - 2*hash->digest_length - 1)
    return FALSE;

  /* This is: emLen - ||M|| - 2hLen - 1  + hLen + 1 + ||M|| =
     emLen - hLen. */
  db_len = emsg_len - hash->digest_length;
  db = ssh_xcalloc(1, db_len);
  /* db_mask = ssh_xmalloc(db_len); */

  /* Compute a hash of the params. */
  context = ssh_xmalloc((*hash->ctxsize)());

  (*hash->reset_context)(context);
  (*hash->update)(context, param, param_len);
  (*hash->final)(context, db);

  /* Add the "01" before the last msg_len bytes. */
  db[db_len - msg_len - 1] = 0x1;

  /* Now throw in the msg. */
  memcpy(db + db_len - msg_len, msg, msg_len);

  /* Generate a random octet string. */
  seed = ssh_xmalloc(hash->digest_length);
  for (i = 0; i < hash->digest_length; i++)
    seed[i] = ssh_random_get_byte();

  /* Now use the MGF1. */
  ssh_rsa_mgf1(hash, seed, hash->digest_length,
               emsg + hash->digest_length, db_len);

  /* Xor. */
  for (i = 0; i < db_len; i++)
    emsg[hash->digest_length + i] ^= db[i];
  memset(db, 0, db_len);

  /* Use MGF1 again. */
  ssh_rsa_mgf1(hash, emsg + hash->digest_length, db_len,
               emsg, hash->digest_length);

  /* Xor the seed. */
  for (i = 0; i < hash->digest_length; i++)
    emsg[i] ^= seed[i];
  memset(seed, 0, hash->digest_length);

  /* Now free the allocated information. */
  ssh_xfree(context);
  ssh_xfree(db);
  ssh_xfree(seed);

  return TRUE;
}

/* OAEP decode using MGF1. */
Boolean ssh_rsa_oaep_decode_with_mgf1(const SshHashDef *hash,
                                      const unsigned char *emsg,
                                      size_t emsg_len,
                                      const unsigned char *param,
                                      size_t param_len,
                                      unsigned char **msg, size_t *msg_len)
{
  void *context;
  unsigned char *seed;
  unsigned char *db;
  unsigned char *phash;
  size_t         db_len, i;
  Boolean        rv = FALSE;

  if (emsg_len < 2*hash->digest_length + 1)
    return FALSE;

  /* Allocate enough working buffers. */
  seed = ssh_xmalloc(hash->digest_length);
  db_len = emsg_len - hash->digest_length;
  db   = ssh_xmalloc(db_len);

  /* Use the mgf. */
  ssh_rsa_mgf1(hash, emsg + hash->digest_length, db_len,
               seed, hash->digest_length);
  /* Now xor. */
  for (i = 0; i < hash->digest_length; i++)
    seed[i] ^= emsg[i];

  /* Use the mgf again. */
  ssh_rsa_mgf1(hash, seed, hash->digest_length,
               db, db_len);
  /* Now xor again. */
  for (i = 0; i < db_len; i++)
    db[i] ^= emsg[hash->digest_length + i];

  /* Compute a hash of the parameters. */
  context = ssh_xmalloc((*hash->ctxsize)());
  phash   = ssh_xmalloc(hash->digest_length);

  (*hash->reset_context)(context);
  (*hash->update)(context, param, param_len);
  (*hash->final)(context, phash);

  /* Do the check. */
  if (memcmp(db, phash, hash->digest_length) != 0)
    goto failed;

  for (i = hash->digest_length; i < db_len; i++)
    {
      if (db[i] != 0)
        {
          if (db[i] != 0x1)
            goto failed;
          break;
        }
    }
  if (i >= db_len)
    goto failed;

  /* Now we must have db[i] == 0x1. */
  *msg_len = db_len - i - 1;
  *msg = ssh_xmalloc(*msg_len);
  memcpy(*msg, db + i + 1, *msg_len);

  rv = TRUE;
failed:
  ssh_xfree(context);
  ssh_xfree(seed);
  ssh_xfree(phash);
  ssh_xfree(db);

  return rv;
}

/* Pad data as pkcs1 suggests. The input_len is needed due the possibility
   that input might contain zeros as most-significant octets which clearly
   makes ssh_mp_get_size return incorrect lengths (for this particular
   need). */

void ssh_rsa_pkcs1_pad_data(SshMPInt output, SshMPInt input,
                            unsigned int input_len,
                            unsigned int tag_number,
                            unsigned int len)
{
  unsigned int i;

  /* Set tag number. */
  ssh_mp_set_ui(output, tag_number);

  /* Check the block type. */
  switch (tag_number)
    {
      /* Block type 0. */
    case 0x0:
      ssh_mp_mul_2exp(output, output, 8 * (len - 2));
      ssh_mp_add(output, output, input);
      break;
      /* Block type 1 (used with signatures). */
    case 0x1:
      for (i = 2; i < len - input_len - 1; i++)
        {
          ssh_mp_mul_2exp(output, output, 8);
          ssh_mp_add_ui(output, output, 0xff);
        }
      ssh_mp_mul_2exp(output, output, 8 * (input_len + 1));
      ssh_mp_add(output, output, input);
      break;
      /* Block type 2 (used with encryption). */
    case 0x2:
      for (i = 2; i < len - input_len - 1; i++)
        {
          unsigned int byte;
          do
            byte = ssh_random_get_byte();
          while (byte == 0);
          ssh_mp_mul_2exp(output, output, 8);
          ssh_mp_add_ui(output, output, byte);
        }
      ssh_mp_mul_2exp(output, output, 8 * (input_len + 1));
      ssh_mp_add(output, output, input);
      break;
    default:
      ssh_fatal("ssh_rsa_pkcs1_pad_data: block type unknown %d.", tag_number);
      break;
    }
}

/* Unpad data as pkcs1 suggests. */

Boolean ssh_rsa_pkcs1_unpad_data(unsigned int tag_number,
                                 SshMPInt input,
                                 unsigned char *output_buffer,
                                 size_t output_buffer_length,
                                 size_t *return_length)
{
  unsigned int i;
  unsigned char *input_buffer;

  /* Linearize to output buffer, which must be long enough. */
  input_buffer = ssh_xmalloc(output_buffer_length);
  ssh_mp_to_buf(input_buffer, output_buffer_length, input);

  /* Check for valid block. */

  if (input_buffer[0] != 0 || input_buffer[1] != tag_number)
    {
      ssh_xfree(input_buffer);
      return FALSE;
    }

  /* Check the block type. */
  switch (tag_number)
    {
      /* Block type 0. */
    case 0x0:
      /* This block type needs further handling at later time... We just
         get it out as is. */
      memcpy(output_buffer, input_buffer + 2, output_buffer_length - 2);
      *return_length = output_buffer_length - 2;
      break;
      /* Block type 1 (used with signatures). */
    case 0x1:
      for (i = 2; i < output_buffer_length; i++)
        {
          if (input_buffer[i] != 0xff)
            break;
        }
      /* Copy. */
      memcpy(output_buffer, input_buffer + i + 1,
             output_buffer_length - i - 1);
      *return_length = output_buffer_length - i - 1;
      break;
      /* Block type 2 (used with encryption). */
    case 0x2:
      for (i = 2; i < output_buffer_length; i++)
        {
          if (input_buffer[i] == 0x0)
            break;
        }

      /* Copy. */
      memcpy(output_buffer, input_buffer + i + 1,
             output_buffer_length - i - 1);
      *return_length = output_buffer_length - i - 1;

      break;
    default:
      ssh_fatal("ssh_rsa_pkcs1_unpad_data: block type unknown %d.",
                tag_number);
      break;
    }

  /* Free temporary buffer. */
  ssh_xfree(input_buffer);

  return TRUE;
}

/* pkcs1 style RSA operations. */

Boolean ssh_rsa_public_key_pkcs1_encrypt(const void *public_key,
                                         const unsigned char *plaintext,
                                         size_t plaintext_len,
                                         unsigned char *ciphertext_buffer,
                                         size_t ciphertext_buffer_len,
                                         size_t *ciphertext_len_return,
                                         const SshHashDef *hash)
{
  const SshRSAPublicKey *pub = public_key;
  SshMPIntStruct t1, t2;

  /* Get size. */
  *ciphertext_len_return = (pub->bits + 7)/8;

  /* Check lengths. */
  if (*ciphertext_len_return < plaintext_len + 1)
    return FALSE;
  if (*ciphertext_len_return > ciphertext_buffer_len)
    return FALSE;

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  ssh_buf_to_mp(&t1, plaintext, plaintext_len);

  /* Pad as pkcs1 suggests. */

  ssh_rsa_pkcs1_pad_data(&t2, &t1, (unsigned int)plaintext_len, 2,
                         *ciphertext_len_return);

  /* Public encrypt. */
  ssh_rsa_public(&t2, &t1, pub);

  /* Linearize again. */
  memset(ciphertext_buffer, 0, *ciphertext_len_return);
  ssh_mp_to_buf(ciphertext_buffer, *ciphertext_len_return, &t1);

  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  return TRUE;
}

Boolean ssh_rsa_private_key_pkcs1_decrypt(const void *private_key,
                                          const unsigned char *ciphertext,
                                          size_t ciphertext_len,
                                          unsigned char *plaintext_buffer,
                                          size_t plaintext_buffer_len,
                                          size_t *plaintext_length_return,
                                          const SshHashDef *hash)
{
  const SshRSAPrivateKey *prv = private_key;
  SshMPIntStruct t1, t2;
  size_t prv_size = (prv->bits + 7)/8;

  /* Check the size of the output buffer.
     XXX This is not entirely satisfactory, the plaintext_buffer_len might
     be desired to be of size less than the modulus...
   */
  if (prv_size > plaintext_buffer_len)
    return FALSE;

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Unlinearize. */
  ssh_buf_to_mp(&t1, ciphertext, ciphertext_len);

  /* Private key operation. */
  if (ssh_rsa_private(&t1, &t2, prv) == SSH_RSA_OP_FAILED)
  {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      return FALSE;
  }

  /* Unpad please. */
  if (ssh_rsa_pkcs1_unpad_data(2, &t2,
                               plaintext_buffer, prv_size,
                               plaintext_length_return) == FALSE)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      return FALSE;
    }

  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  return TRUE;
}

/* None encrypt and decrypt */
Boolean ssh_rsa_public_key_none_encrypt(const void *public_key,
                                        const unsigned char *plaintext,
                                        size_t plaintext_len,
                                        unsigned char *ciphertext_buffer,
                                        size_t ciphertext_buffer_len,
                                        size_t *ciphertext_len_return,
                                        const SshHashDef *hash)
{
  const SshRSAPublicKey *pub = public_key;
  SshMPIntStruct t1, t2;

  /* Get size. */
  *ciphertext_len_return = (pub->bits + 7)/8;

  /* Check lengths. */
  if (*ciphertext_len_return != plaintext_len)
    return FALSE;
  if (*ciphertext_len_return > ciphertext_buffer_len)
    return FALSE;

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  ssh_buf_to_mp(&t1, plaintext, plaintext_len);

  /* Public encrypt. */
  ssh_rsa_public(&t1, &t2, pub);

  /* Linearize again. */
  memset(ciphertext_buffer, 0, *ciphertext_len_return);
  ssh_mp_to_buf(ciphertext_buffer, *ciphertext_len_return, &t2);

  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  return TRUE;
}

Boolean ssh_rsa_private_key_none_decrypt(const void *private_key,
                                         const unsigned char *ciphertext,
                                         size_t ciphertext_len,
                                         unsigned char *plaintext_buffer,
                                         size_t plaintext_buffer_len,
                                         size_t *plaintext_length_return,
                                         const SshHashDef *hash)
{
  const SshRSAPrivateKey *prv = private_key;
  SshMPIntStruct t1, t2;

  *plaintext_length_return = (prv->bits + 7)/8;

  /* Check the size of the output buffer.
     XXX This is not entirely satisfactory, the plaintext_buffer_len might
     be desired to be of size less than the modulus...
   */
  if (*plaintext_length_return > plaintext_buffer_len)
    return FALSE;

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Unlinearize. */
  ssh_buf_to_mp(&t1, ciphertext, ciphertext_len);

  /* Private key operation. */
  if (ssh_rsa_private(&t1, &t2, prv) == SSH_RSA_OP_FAILED)
  {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      return FALSE;
  }

  /* Linearize again. */
  ssh_mp_to_buf(plaintext_buffer, *plaintext_length_return, &t2);

  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  return TRUE;
}

/* This function is specific to PKCS-1v2 and might need revision
   after possible updates. However, there is little we can help
   here. */
unsigned char *ssh_rsa_pkcs1v2_default_explicit_param(const SshHashDef *hash,
                                                      size_t *param_len)
{
  SshAsn1Status  status;
  SshAsn1Context context;
  SshAsn1Node    node;
  unsigned char *param;

  /* Initialize the ASN.1 library. */
  context = ssh_asn1_init();

  /* Useful OIDs. */
#define SSH_RSA_MGF1_OID    "1.2.840.113549.1.1.8"
#define SSH_RSA_PSOURCE_OID "1.2.840.113549.1.1.9"

  status = ssh_asn1_create_node(context, &node,
                                "(sequence ()"
                                "  (sequence ()"
                                "    (object-identifier ())"
                                "    (null ()))"                 /* Hash */
                                "  (sequence ()"
                                "    (object-identifier ())"     /* MGF */
                                "    (sequence ()"
                                "      (object-identifier ())"
                                "      (null ())))"              /* Hash */
                                "  (sequence ()"
                                "    (object-identifier ())"     /* psource */
                                "    (octet-string ())))",
                                hash->asn1_oid,
                                SSH_RSA_MGF1_OID,
                                hash->asn1_oid,
                                SSH_RSA_PSOURCE_OID,
                                NULL, 0);
  if (status != SSH_ASN1_STATUS_OK)
    {
      ssh_asn1_free(context);
      return NULL;
    }

  /* Now encode the node. */

  status = ssh_asn1_encode_node(context, node);
  if (status != SSH_ASN1_STATUS_OK)
    {
      ssh_asn1_free(context);
      return NULL;
    }

  status =
    ssh_asn1_node_get_data(node, &param, param_len);
  if (status != SSH_ASN1_STATUS_OK)
    {
      ssh_asn1_free(context);
      return NULL;
    }

  ssh_asn1_free(context);

  return param;
}


Boolean ssh_rsa_public_key_pkcs1v2_encrypt(const void *public_key,
                                           const unsigned char *plaintext,
                                           size_t plaintext_len,
                                           unsigned char *ciphertext_buffer,
                                           size_t ciphertext_buffer_len,
                                           size_t *ciphertext_len_return,
                                           const SshHashDef *hash)
{
  const SshRSAPublicKey *pub = public_key;
  SshMPIntStruct t1, t2;
  unsigned char *emsg, *param;
  size_t param_len;

  /* Get size. */
  *ciphertext_len_return = (pub->bits + 7)/8;

  /* Check lengths. */
  if (*ciphertext_len_return - 2 - 2*hash->digest_length < plaintext_len)
    return FALSE;
  if (*ciphertext_len_return > ciphertext_buffer_len)
    return FALSE;

  /* Create the default explicit parameters. */
  param = ssh_rsa_pkcs1v2_default_explicit_param(hash, &param_len);
  if (param == NULL)
    return FALSE;

  emsg = ssh_xmalloc(*ciphertext_len_return - 1);

  /* Now apply the OAEP encoding. */
  if (ssh_rsa_oaep_encode_with_mgf1(hash,
                                    plaintext, plaintext_len,
                                    param, param_len,
                                    emsg, *ciphertext_len_return - 1) == FALSE)
    {
      ssh_xfree(param);
      ssh_xfree(emsg);
      return FALSE;
    }

  ssh_xfree(param);

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  ssh_buf_to_mp(&t1, emsg, *ciphertext_len_return - 1);

  /* Public encrypt. */
  ssh_rsa_public(&t1, &t2, pub);

  /* Linearize again. */
  ssh_mp_to_buf(ciphertext_buffer, *ciphertext_len_return, &t2);

  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  ssh_xfree(emsg);

  return TRUE;
}

Boolean ssh_rsa_private_key_pkcs1v2_decrypt(const void *private_key,
                                            const unsigned char *ciphertext,
                                            size_t ciphertext_len,
                                            unsigned char *plaintext_buffer,
                                            size_t plaintext_buffer_len,
                                            size_t *plaintext_length_return,
                                            const SshHashDef *hash)
{
  const SshRSAPrivateKey *prv = private_key;
  SshMPIntStruct t1, t2;
  size_t prv_size = (prv->bits + 7)/8;
  unsigned char *param, *emsg, *msg;
  size_t param_len, msg_len;

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Unlinearize. */
  ssh_buf_to_mp(&t1, ciphertext, ciphertext_len);

  /* Private key operation. */
  if (ssh_rsa_private(&t1, &t2, prv) == SSH_RSA_OP_FAILED)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      return FALSE;
    }

  /* Find params. */
  param = ssh_rsa_pkcs1v2_default_explicit_param(hash, &param_len);
  if (param == NULL)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      return FALSE;
    }

  /* Convert the decoded integer back to suitable octet string. */
  emsg = ssh_xmalloc(prv_size - 1);
  ssh_mp_to_buf(emsg, prv_size - 1, &t2);

  /* Apply the OAEP decoding. */
  if (ssh_rsa_oaep_decode_with_mgf1(hash,
                                    emsg, prv_size - 1,
                                    param, param_len,
                                    &msg, &msg_len) == FALSE)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      ssh_xfree(emsg);
      ssh_xfree(param);
      return FALSE;
    }

  /* Copy to output. */
  if (plaintext_buffer_len < msg_len)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      ssh_xfree(emsg);
      ssh_xfree(param);
      ssh_xfree(msg);
      return FALSE;
    }

  memcpy(plaintext_buffer, msg, msg_len);
  *plaintext_length_return = msg_len;

  /* Free allocated data. */
  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);
  ssh_xfree(emsg);
  ssh_xfree(param);
  ssh_xfree(msg);

  return TRUE;
}

Boolean ssh_rsa_public_key_pkcs1_verify(const void *public_key,
                                        const unsigned char *signature,
                                        size_t signature_len,
                                        Boolean need_hashing,
                                        const unsigned char *data,
                                        size_t data_len,
                                        const SshHashDef *hash)
{
  SshAsn1Context asn1_context;
  SshAsn1Status status;
  SshAsn1Tree tree;
  unsigned char *ber_buf;
  size_t ber_buf_len;
  const SshRSAPublicKey *pub = public_key;
  SshMPIntStruct t1, t2;
  char *oid = NULL;
  unsigned char *temp_buf = NULL;
  size_t return_len;
  Boolean rv = FALSE;
  void *context;
  unsigned char *digest;

  /* Verify data length if needs hashing. */
  if (need_hashing == FALSE && hash->digest_length != data_len)
    return rv;

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Unlinearize. */
  ssh_buf_to_mp(&t1, signature, signature_len);
  if (ssh_mp_cmp(&t1, &pub->n) >= 0 ||
      ssh_mp_cmp_ui(&t1, 0) <= 0)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      goto failed;
    }

  /* Allocate temporary buffer. */
  ber_buf_len = (pub->bits + 7)/8;
  ber_buf = ssh_xmalloc(ber_buf_len);

  /* Retrieve the original for of the signature. */
  ssh_rsa_public(&t1, &t2, pub);

  /* Unpad. */
  if (ssh_rsa_pkcs1_unpad_data(1, &t2,
                               ber_buf, ber_buf_len,
                               &return_len) == FALSE)
    {
      ssh_xfree(ber_buf);
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      goto failed;
    }

  /* Clear multiple precision integers. */
  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  /* Now we have the "pleasant" job of decoding ASN.1 BER coded blob. */
  asn1_context = ssh_asn1_init();

  status = ssh_asn1_decode(asn1_context, ber_buf, return_len, &tree);

  /* Free BER buffer. */
  ssh_xfree(ber_buf);

  if (status != SSH_ASN1_STATUS_OK)
    {
      ssh_asn1_free(asn1_context);
      goto failed;
    }

  context = NULL;
  status = ssh_asn1_read_tree(context, tree,
                              "(sequence ()"
                              "  (sequence ()"
                              "    (object-identifier ())"
                              "    (null ()))"
                              "  (octet-string ()))",
                              &oid,
                              &temp_buf, &return_len);

  /* Free asn1 context. */
  ssh_asn1_free(asn1_context);
  if (status != SSH_ASN1_STATUS_OK)
    {
      if (oid) ssh_xfree(oid);
      if (temp_buf) ssh_xfree(temp_buf);
      goto failed;
    }

  /* Hash if needed. */
  if (need_hashing)
    {
      /* Compute hash digest. */
      context = ssh_xmalloc((*hash->ctxsize)());
      digest = ssh_xmalloc(hash->digest_length);

      /* Compute hash of buffer. */
      (*hash->reset_context)(context);
      (*hash->update)(context, data, data_len);
      (*hash->final)(context, digest);
      ssh_xfree(context);
    }
  else
    digest = (unsigned char *)data;

  /* Compare. */
  if (hash->digest_length != return_len)
    {
      ssh_xfree(oid);
      ssh_xfree(temp_buf);
      if (need_hashing)
        ssh_xfree(digest);
      goto failed;
    }
  if (memcmp(digest, temp_buf, hash->digest_length) == 0 &&
      strcmp(hash->asn1_oid, oid) == 0)
    rv = TRUE;

  ssh_xfree(oid);
  if (need_hashing)
    ssh_xfree(digest);
  ssh_xfree(temp_buf);
failed:

  return rv;
}

Boolean ssh_rsa_private_key_pkcs1_sign(const void *private_key,
                                       Boolean need_hashing,
                                       const unsigned char *data,
                                       size_t data_len,
                                       unsigned char *signature_buffer,
                                       size_t signature_buffer_len,
                                       size_t *signature_length_return,
                                       const SshHashDef *hash)
{
  const SshRSAPrivateKey *prv = private_key;
  SshMPIntStruct t1, t2;
  void *context;
  SshAsn1Context asn1_context;
  SshAsn1Tree tree;
  SshAsn1Status status;
  unsigned char *ber_data;
  size_t ber_data_len;
  unsigned char *digest;
  SshRSAOpStatus rsa_op_status;

  if (need_hashing)
    {
      /* Do the pkcs-1 magic. */
      context = ssh_xmalloc((*hash->ctxsize)());
      digest = ssh_xmalloc(hash->digest_length);

      /* Hash the given data. */
      (*hash->reset_context)(context);
      (*hash->update)(context, data, data_len);
      (*hash->final)(context, digest);
      ssh_xfree(context);
    }
  else
    {
      digest = (unsigned char *)data;
      if (data_len != hash->digest_length)
        return FALSE;
    }

  /* Make a ASN.1/BER encoded string. */
  asn1_context = ssh_asn1_init();

  /* Build simple ASN.1 tree which contains just basic sequences and
     object identifiers and octet strings. If speed is mandatory we
     could precompute parts of the der encoding and place appropriate
     lengths here. */
  status = ssh_asn1_create_tree(asn1_context, &tree,
                                "(sequence ()"
                                "  (sequence ()"
                                "    (object-identifier ())"
                                "    (null ()))"
                                "  (octet-string ()))",
                                hash->asn1_oid,
                                digest, hash->digest_length);

  if (need_hashing)
    ssh_xfree(digest);

  /* Verify that the tree was correctly reconstructed. */
  if (status != SSH_ASN1_STATUS_OK)
    {
      ssh_asn1_free(asn1_context);
      return FALSE;
    }

  /* Do the encoding. */
  status = ssh_asn1_encode(asn1_context, tree);
  if (status != SSH_ASN1_STATUS_OK)
    {
      ssh_asn1_free(asn1_context);
      return FALSE;
    }

  /* Get BER encoded blob. */
  ssh_asn1_get_data(tree, &ber_data, &ber_data_len);
  /* Free ASN.1 context. */
  ssh_asn1_free(asn1_context);

  /* Compute signature. */

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Unlinearize. */
  ssh_buf_to_mp(&t1, ber_data, ber_data_len);

  /* Free the ber data allocate by ssh_asn1_get_data. */
  ssh_xfree(ber_data);

  *signature_length_return = (prv->bits + 7)/8;

  /* Pad data as pkcs1 suggests. */
  if (*signature_length_return < ber_data_len + 1)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      return FALSE;
    }

  ssh_rsa_pkcs1_pad_data(&t2, &t1, (unsigned int)ber_data_len,
                         1, *signature_length_return);

  /* Private key operation. */
  rsa_op_status = ssh_rsa_private(&t2, &t1, prv);

  if (rsa_op_status == SSH_RSA_OP_FAILED)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      return FALSE;
    }

  /* Linearize. */
  ssh_mp_to_buf(signature_buffer, *signature_length_return, &t1);

  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  return TRUE;
}

Boolean ssh_rsa_public_key_pkcs1_none_verify(const void *public_key,
                                             const unsigned char *signature,
                                             size_t signature_len,
                                             Boolean need_hashing,
                                             const unsigned char *data,
                                             size_t data_len,
                                             const SshHashDef *hash)
{
  const SshRSAPublicKey *pub = public_key;
  SshMPIntStruct t1, t2;
  unsigned char *buf;
  size_t return_len, buf_len;
  Boolean rv = FALSE;

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Unlinearize. */
  ssh_buf_to_mp(&t1, signature, signature_len);
  if (ssh_mp_cmp(&t1, &pub->n) >= 0 ||
      ssh_mp_cmp_ui(&t1, 0) <= 0)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      goto failed;
    }

  /* Allocate temporary buffer. */
  buf_len = (pub->bits + 7)/8;
  buf = ssh_xmalloc(buf_len);

  /* Retrieve the original for of the signature. */
  ssh_rsa_public(&t1, &t2, pub);

  /* Unpad. */
  if (ssh_rsa_pkcs1_unpad_data(1, &t2,
                               buf, buf_len,
                               &return_len) == FALSE)
    {
      ssh_xfree(buf);
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      goto failed;
    }

  /* Clear multiple precision integers. */
  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  /* Compare. */
  if (data_len != return_len || memcmp(data, buf, data_len) != 0)
    {
      ssh_xfree(buf);
      goto failed;
    }

  ssh_xfree(buf);
  rv = TRUE;
failed:

  return rv;
}

Boolean ssh_rsa_private_key_pkcs1_none_sign(const void *private_key,
                                            Boolean need_hashing,
                                            const unsigned char *data,
                                            size_t data_len,
                                            unsigned char *signature_buffer,
                                            size_t signature_buffer_len,
                                            size_t *signature_length_return,
                                            const SshHashDef *hash)
{
  const SshRSAPrivateKey *prv = private_key;
  SshMPIntStruct t1, t2;
  int rsa_status;

  *signature_length_return = (prv->bits + 7)/8;

  if (*signature_length_return < data_len + 1)
    return FALSE;
  if (*signature_length_return > signature_buffer_len)
    return FALSE;

  /* Compute signature. */

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Unlinearize. */
  ssh_buf_to_mp(&t1, data, data_len);

  /* Pad data as pkcs1 suggests. */
  ssh_rsa_pkcs1_pad_data(&t2, &t1, (unsigned int)data_len,
                         1, *signature_length_return);

  /* Private key operation. */
  rsa_status = ssh_rsa_private(&t2, &t1, prv);
  if (rsa_status == SSH_RSA_OP_FAILED)
  {
    ssh_mp_clear(&t1);
    ssh_mp_clear(&t2);

    return FALSE;
  }
  if (rsa_status == SSH_RSA_OP_OK)
    /* Linearize. */
    ssh_mp_to_buf(signature_buffer, *signature_length_return, &t1);

  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  return TRUE;
}

Boolean ssh_rsa_public_key_none_decrypt(const void *public_key,
                                        const unsigned char *ciphertext,
                                        size_t ciphertext_len,
                                        unsigned char *plaintext,
                                        size_t plaintext_len,
                                        size_t *plaintext_length_return)
{
  const SshRSAPublicKey *pub = public_key;
  SshMPIntStruct t1, t2;

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Unlinearize. */
  ssh_buf_to_mp(&t1, ciphertext, ciphertext_len);
  if (ssh_mp_cmp(&t1, &pub->n) >= 0 ||
      ssh_mp_cmp_ui(&t1, 0) <= 0)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      return FALSE;
    }

  /* Allocate temporary buffer. */
  *plaintext_length_return = (pub->bits + 7)/8;
  if (plaintext_len < *plaintext_length_return)
    return FALSE;

  /* Retrieve the original for of the ciphertext. */
  ssh_rsa_public(&t1, &t2, pub);

  /* Unlinearize. */
  ssh_mp_to_buf(plaintext, *plaintext_length_return, &t2);

  /* Clear multiple precision integers. */
  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);
  return TRUE;
}

Boolean ssh_rsa_private_key_none_encrypt(const void *private_key,
                                         const unsigned char *data,
                                         size_t data_len,
                                         unsigned char *signature_buffer,
                                         size_t signature_buffer_len,
                                         size_t *signature_length_return)
{
  const SshRSAPrivateKey *prv = private_key;
  SshMPIntStruct t1, t2;
  int rsa_status;

  *signature_length_return = (prv->bits + 7)/8;

  if (*signature_length_return != data_len)
    return FALSE;
  if (*signature_length_return > signature_buffer_len)
    return FALSE;

  /* Compute signature. */

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Unlinearize. */
  ssh_buf_to_mp(&t1, data, data_len);

  /* Private key operation. */
  rsa_status = ssh_rsa_private(&t1, &t2, prv);
  if (rsa_status == SSH_RSA_OP_FAILED)
  {
    ssh_mp_clear(&t1);
    ssh_mp_clear(&t2);

    return FALSE;
  }
  if (rsa_status == SSH_RSA_OP_OK)
    /* Linearize. */
    ssh_mp_to_buf(signature_buffer, *signature_length_return, &t2);

  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  return TRUE;
}

/* The routines using the RGF's. */

Boolean ssh_rsa_public_key_encrypt(const void *public_key,
                                   const unsigned char *plaintext,
                                   size_t plaintext_len,
                                   unsigned char *ciphertext_buffer,
                                   size_t ciphertext_buffer_len,
                                   size_t *ciphertext_len_return,
                                   const SshRGFDef *rgf_def)
{
  const SshRSAPublicKey *pub = public_key;
  SshMPIntStruct t1, t2;
  unsigned char *buf;

  /* Get size. */
  *ciphertext_len_return = (pub->bits + 7)/8;

  /* Check lengths. */
  if (*ciphertext_len_return < plaintext_len)
    return FALSE;
  if (*ciphertext_len_return > ciphertext_buffer_len)
    return FALSE;

  /* Allocate large enough buffer. */
  buf = ssh_xmalloc(*ciphertext_len_return);

  /* Now run the RGF. */
  if ((*rgf_def->rgf_encrypt)(plaintext, plaintext_len,
                              rgf_def,
                              buf, *ciphertext_len_return) != SSH_RGF_OK)
    {
      ssh_xfree(buf);
      return FALSE;
    }

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Convert to integer. */
  ssh_buf_to_mp(&t2, buf, *ciphertext_len_return);
  ssh_xfree(buf);

  /* Public encrypt. */
  ssh_rsa_public(&t2, &t1, pub);

  /* Linearize again. */
  ssh_mp_to_buf(ciphertext_buffer, *ciphertext_len_return, &t1);

  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  return TRUE;
}

Boolean ssh_rsa_private_key_decrypt(const void *private_key,
                                    const unsigned char *ciphertext,
                                    size_t ciphertext_len,
                                    unsigned char *plaintext_buffer,
                                    size_t plaintext_buffer_len,
                                    size_t *plaintext_length_return,
                                    const SshRGFDef *rgf_def)
{
  const SshRSAPrivateKey *prv = private_key;
  SshMPIntStruct t1, t2;
  size_t prv_size = (prv->bits + 7)/8;
  unsigned char *decrypted_msg, *output_msg;
  size_t output_msg_len;

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Unlinearize. */
  ssh_buf_to_mp(&t1, ciphertext, ciphertext_len);

  /* Private key operation. */
  if (ssh_rsa_private(&t1, &t2, prv) == SSH_RSA_OP_FAILED)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      return FALSE;
    }

  /* Linearize. */
  decrypted_msg = ssh_xmalloc(prv_size);
  ssh_mp_to_buf(decrypted_msg, prv_size, &t2);

  if ((*rgf_def->rgf_decrypt)(decrypted_msg, prv_size,
                              prv_size,
                              rgf_def,
                              &output_msg, &output_msg_len) != SSH_RGF_OK)
    {
      ssh_xfree(decrypted_msg);
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      return FALSE;
    }

  ssh_xfree(decrypted_msg);
  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  /* Now check the output size. */
  if (output_msg_len > plaintext_buffer_len)
    {
      ssh_xfree(output_msg);
      return FALSE;
    }

  memcpy(plaintext_buffer, output_msg, output_msg_len);
  *plaintext_length_return = output_msg_len;
  ssh_xfree(output_msg);

  return TRUE;
}

Boolean ssh_rsa_public_key_verify(const void *public_key,
                                  const unsigned char *signature,
                                  size_t signature_len,
                                  SshRGFHash       hash,
                                  const SshRGFDef *rgf_def)
{
  const SshRSAPublicKey *pub = public_key;
  SshMPIntStruct t1, t2;
  unsigned char *buf, *output_buf;
  size_t buf_len, output_buf_len;
  Boolean rv = FALSE;

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Unlinearize. */
  ssh_buf_to_mp(&t1, signature, signature_len);
  if (ssh_mp_cmp(&t1, &pub->n) >= 0 ||
      ssh_mp_cmp_ui(&t1, 0) <= 0)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      ssh_rgf_hash_free(hash);
      goto failed;
    }

  /* Allocate temporary buffer. */
  buf_len = (pub->bits + 7)/8;
  buf = ssh_xmalloc(buf_len);

  /* Retrieve the original for of the signature. */
  ssh_rsa_public(&t1, &t2, pub);

  /* Unlinearize. */
  ssh_mp_to_buf(buf, buf_len, &t2);

  /* Clear multiple precision integers. */
  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  /* Do the verification. */
  if ((*rgf_def->rgf_verify)(hash, buf, buf_len, buf_len,
                             &output_buf, &output_buf_len) != SSH_RGF_OK)
    {
      ssh_xfree(buf);
      goto failed;
    }
  ssh_xfree(buf);
  if (output_buf != NULL)
    ssh_xfree(output_buf);
  rv = TRUE;
failed:

  return rv;

}

Boolean ssh_rsa_private_key_sign(const void *private_key,
                                 SshRGFHash    hash,
                                 unsigned char *signature_buffer,
                                 size_t signature_buffer_len,
                                 size_t *signature_length_return,
                                 const SshRGFDef *rgf_def)
{
  const SshRSAPrivateKey *prv = private_key;
  unsigned char *output_msg;
  SshMPIntStruct t1, t2;
  int rsa_status;

  *signature_length_return = (prv->bits + 7)/8;

  if (*signature_length_return > signature_buffer_len)
    {
      ssh_rgf_hash_free(hash);
      return FALSE;
    }

  output_msg = ssh_xmalloc(*signature_length_return);

  /* Build the to be signed data. */
  if ((*rgf_def->rgf_sign)(hash, output_msg,
                           *signature_length_return) != SSH_RGF_OK)
    {
      ssh_xfree(output_msg);
      return FALSE;
    }

  /* Compute signature. */

  ssh_mp_init(&t1);
  ssh_mp_init(&t2);

  /* Unlinearize. */
  ssh_buf_to_mp(&t1, output_msg, *signature_length_return);
  ssh_xfree(output_msg);

  /* Private key operation. */
  rsa_status = ssh_rsa_private(&t1, &t2, prv);
  if (rsa_status == SSH_RSA_OP_FAILED)
    {
      ssh_mp_clear(&t1);
      ssh_mp_clear(&t2);
      return FALSE;
    }
  /* Linearize. */
  ssh_mp_to_buf(signature_buffer, *signature_length_return, &t2);

  ssh_mp_clear(&t1);
  ssh_mp_clear(&t2);

  return TRUE;
}

/* Compute sizes needed in each operation. */

size_t ssh_rsa_public_key_max_encrypt_input_len(const void *public_key)
{
  const SshRSAPublicKey *pub = public_key;
  size_t len = ((pub->bits + 7)/8 - 3 - SSH_RSA_MINIMUM_PADDING);
  if (len > 0 && len < SSH_RSA_MAX_BYTES)
    return len;
  return 0;
}

/* Note: this should be changed to take the hash function as an argument.
   Indeed, even more nicely take the MGF as an argument. */
size_t ssh_rsa_public_key_max_oaep_encrypt_input_len(const void *public_key)
{
  const SshRSAPublicKey *pub = public_key;
  size_t len = ((pub->bits + 7)/8 - 2 - 2*SSH_MAX_HASH_DIGEST_LENGTH);
  if (len > 0 && len < SSH_RSA_MAX_BYTES)
    return len;
  return 0;
}

size_t ssh_rsa_public_key_max_none_encrypt_input_len(const void *public_key)
{
  const SshRSAPublicKey *pub = public_key;
  size_t len = ((pub->bits + 7)/8);
  if (len > 0 && len < SSH_RSA_MAX_BYTES)
    return len;
  return 0;
}

size_t ssh_rsa_private_key_max_decrypt_input_len(const void *private_key)
{
  const SshRSAPrivateKey *prv = private_key;
  return (prv->bits + 7)/8;
}

size_t ssh_rsa_private_key_max_signature_input_len(const void *private_key)
{
  return (size_t)-1;
}

size_t
ssh_rsa_private_key_max_signature_unhash_input_len(const void *private_key)
{
  const SshRSAPrivateKey *prv = private_key;
  size_t len = ((prv->bits + 7)/8 - 3 - SSH_RSA_MINIMUM_PADDING);
  if (len > 0 && len < SSH_RSA_MAX_BYTES)
    return len;
  return 0;
}

size_t ssh_rsa_public_key_max_encrypt_output_len(const void *public_key)
{
  const SshRSAPublicKey *pub = public_key;
  return (pub->bits + 7)/8;
}

size_t ssh_rsa_private_key_max_decrypt_output_len(const void *private_key)
{
  const SshRSAPrivateKey *prv = private_key;
  return (prv->bits + 7)/8;
}

size_t ssh_rsa_private_key_max_signature_output_len(const void *private_key)
{
  const SshRSAPrivateKey *prv = private_key;
  return (prv->bits + 7)/8;
}

/* Importing and exporting private keys. */

Boolean ssh_rsa_private_key_export(const void *private_key,
                                   unsigned char **buf,
                                   size_t *length_return)
{
  const SshRSAPrivateKey *prv = private_key;

  /* Linearize. */
  *length_return =
    ssh_encode_array_alloc(buf,
                     SSH_FORMAT_MP_INT, &prv->e,
                     SSH_FORMAT_MP_INT, &prv->d,
                     SSH_FORMAT_MP_INT, &prv->n,
                     SSH_FORMAT_MP_INT, &prv->u,
                     SSH_FORMAT_MP_INT, &prv->p,
                     SSH_FORMAT_MP_INT, &prv->q,
                     SSH_FORMAT_END);

  return TRUE;
}

Boolean ssh_rsa_private_key_import(const unsigned char *buf,
                                   size_t len,
                                   void **private_key)
{
  SshRSAPrivateKey *prv = ssh_xmalloc(sizeof(*prv));

  /* Initialize. */
  ssh_mp_init(&prv->n);
  ssh_mp_init(&prv->e);
  ssh_mp_init(&prv->d);
  ssh_mp_init(&prv->u);
  ssh_mp_init(&prv->p);
  ssh_mp_init(&prv->q);

  /* Unlinearize. */
  if (ssh_decode_array(buf, len,
                       SSH_FORMAT_MP_INT, &prv->e,
                       SSH_FORMAT_MP_INT, &prv->d,
                       SSH_FORMAT_MP_INT, &prv->n,
                       SSH_FORMAT_MP_INT, &prv->u,
                       SSH_FORMAT_MP_INT, &prv->p,
                       SSH_FORMAT_MP_INT, &prv->q,
                       SSH_FORMAT_END) == 0)
    {
      ssh_mp_clear(&prv->n);
      ssh_mp_clear(&prv->e);
      ssh_mp_clear(&prv->u);
      ssh_mp_clear(&prv->d);
      ssh_mp_clear(&prv->p);
      ssh_mp_clear(&prv->q);
      ssh_xfree(prv);
      return FALSE;
    }

  prv->bits = ssh_mp_bit_size(&prv->n);

  *private_key = (void *)prv;

  return TRUE;
}

Boolean ssh_rsa_public_key_import(const unsigned char *buf,
                                  size_t len,
                                  void **public_key)
{
  SshRSAPublicKey *pub = ssh_xmalloc(sizeof(*pub));

  ssh_mp_init(&pub->n);
  ssh_mp_init(&pub->e);

  if (ssh_decode_array(buf, len,
                       SSH_FORMAT_MP_INT, &pub->e,
                       SSH_FORMAT_MP_INT, &pub->n,
                       SSH_FORMAT_END) == 0)
    {
      ssh_mp_clear(&pub->n);
      ssh_mp_clear(&pub->e);
      ssh_xfree(&pub);
      return FALSE;
    }

  pub->bits = ssh_mp_bit_size(&pub->n);

  *public_key = pub;

  return TRUE;
}

Boolean ssh_rsa_public_key_export(const void *public_key,
                                  unsigned char **buf,
                                  size_t *length_return)
{
  const SshRSAPublicKey *pub = public_key;

  *length_return = ssh_encode_array_alloc(buf,
                                    SSH_FORMAT_MP_INT, &pub->e,
                                    SSH_FORMAT_MP_INT, &pub->n,
                                    SSH_FORMAT_END);

  return TRUE;
}

#endif /* WITH_RSA */

/* rsa.c */
