/*
  sshthread.h

  Author: Jukka Aittokallio <jai@ssh.com>
          Vesa Suontama <vsuontam@ssh.com>

  Copyright (C) 2000 SSH Communications Security Oy, Espoo, Finland
  All rights reserved.

  Generic interface to threads. Some systems may offer some
  synchronous function interfaces (example of this is the
  smart card API in Microsoft Windows). Because all the time
  consuming functions within SSH libraries should be asynchronous,
  the only way is to "threadify" those synchronous calls. Be aware that
  most of the modules in SSH library are not thread safe, so
  threads are to be used very carefully. Generally, use threads
  only to make external synchronous APIs asynchronous.
*/
#ifndef SSH_THREAD_H
#define SSH_THREAD_H

/* Handle to a thread. You can obtain handles using 
   ssh_thread_create or ssh_thread_pool_get_thread. */
typedef struct SshThreadRec *SshThread;

/* handle to a thread pool. Thread pool is an object that manages creation 
   and destruction of threads in such a way that subsequent calls to
   ssh_thread_pool_get_thread may reuse threads that are already finished. 
   Usually the operating systems thread fork is a heavy operation, and reusing 
   already forked threads saves system resources. 

   The pool is created with ssh_thread_pool_create. */
typedef struct SshThreadPoolRec *SshThreadPool;

/* When you call ssh_thread_execute, it starts executing this type of 
   function from it. Do NOT call ssh_thread_destroy or 
   ssh_thread_pool_destroy in this function for this thread. */
typedef void (*SshThreadFuncCB)(void *context);

/* Called when the thread has finished executing the SshThredFuncCB */
typedef void (*SshThreadDoneCB)(void *context);

/* Creates a thread without a pool. You must call ssh_thread_destroy
   when you are done with a thread. */
SshThread
ssh_thread_create(void);

/* Waits for thread to finish its execution and destroys it. Do NOT
   call this function for pooled threads. Pool will take care of
   thread destruction for the threads obtained from pool. */
void
ssh_thread_destroy(SshThread thread);

/* Waits for a certain period the thread to finish executing the
   function SshThreadFuncCB given as an argument to ssh_thread
   execute function. */
Boolean
ssh_thread_wait(SshThread thread, SshUInt32 timeout_ms);

/* Sleeps the thread. This should only be called from a forked thread to 
   suspend the thread for certain amount of time. */
void
ssh_thread_sleep(SshUInt32 wait_ms);

/* Starts executing a function in a separate thread. Calls done_cb
   when thread has finished executing the 'func' function. */
Boolean
ssh_thread_execute(SshThread thread,
                   SshThreadFuncCB func,
                   SshThreadDoneCB done_cb,
                   void *context);

/* Thread pool creation parameters. These parameters define how the thread 
   pool behaves. Pass a pointer to this struct to ssh_thread_pool_create
   if default values are not suitable. */
typedef struct SshThreadPoolParamsRec 
{
  /* The minimum numbers of threads on the pool. This number of threads is 
     created at the creation time of the pool. Setting this argument to nonzero 
     is reasonalbe only in situations where performance is critical. */
  SshUInt32 min_threads;

  /* Maximum number of threads in the pool. If set to zero, there is no maximum 
     number for threads in the pool (except the OS limit, which is dependent of 
     the OS and thread library version) */
  SshUInt32 max_threads;
} SshThreadPoolParamsStruct, *SshThreadPoolParams;

/* Creates a thread pool. The params define how the pool behaves. If params is 
   NULL, all the arguments get their default values. */
SshThreadPool
ssh_thread_pool_create(SshThreadPoolParams params);

/* Waits for all of the threads in the pool to finish their
   execution and destroys the pool. */
void
ssh_thread_pool_destroy(SshThreadPool pool);

/* Waits for a certain period of all the threads to finish
   their execution. Returns FALSE, if one or more threads didn't
   finish in the given timeout period. */
Boolean
ssh_thread_pool_wait(SshThreadPool pool, SshUInt32 timeout_ms);

/* Returns a new or 'reused' thread from the pool, if the maximum number of 
   threads has not been reached. Returns NULL if there are no more threads 
   available in the pool. Note that you may get pointer to the same thread 
   by subsequent calls to this function, if a thread has finished 
   its callback function before the next call to ssh_thread_pool_get_thread is
   made. 
   
   Do not call ssh_thread_destroy for the threads returned from the pool. */
SshThread
ssh_thread_pool_get_thread(SshThreadPool pool);

#endif /* SSH_THREAD_H */
